@@ -62,7 +62,13 @@ class View:
         try:
             return self.tree.keys()
         except AttributeError:
-            return range(len(self.tree))
+            if isinstance(self.tree, list):
+                try:
+                    return [item[NAME] for item in self.tree]
+                except (TypeError, KeyError):
+                    pass
+
+        return range(len(self.tree))
 
     def cp(self, *references: Key):
         *sources, dst = references
