@@ -1,12 +1,12 @@
+from collections import defaultdict
 from copy import deepcopy
 from dataclasses import dataclass, field
 from types import TracebackType
 from typing import Dict, List
-import logging
 import sys
 
 from mash.doc_inference import generate_docs
-from mash.io_util import log
+from mash.io_util import log, log_once
 from mash import util
 
 # this data is impacts by both the classes Function and Shell, hence it should be global
@@ -17,6 +17,10 @@ last_exception: Exception = None
 last_traceback: TracebackType = None
 
 
+LAST_RESULTS = '_last_results'
+LAST_RESULTS_INDEX = '_last_results_index'
+
+
 class ShellFunction:
     def __init__(self, func, func_name=None, synopsis: str = None, args: Dict[str, str] = None, doc: str = None) -> None:
         try:
@@ -29,7 +33,7 @@ class ShellFunction:
             func = deepcopy(func)
         except TypeError as e:
             name = getattr(func, '__name__', 'func')
-            logging.debug(f'Cannot deepcopy `{name}`: {e.args[0]}')
+            log_once(f'Cannot deepcopy `{name}`: {e.args[0]}')
 
         self.func = func
 
@@ -65,3 +69,7 @@ class InlineFunction:
     multiline: bool = False
     line_indent: str = None
     inner: List[str] = field(default_factory=list)
+
+
+def scope() -> dict:
+    return defaultdict(dict)
