@@ -1,10 +1,12 @@
 from functools import partial
+from logging import debug
 
 from mash.filesystem.filesystem import FileSystem, OPTIONS, Option
 from mash.filesystem.discoverable import Discoverable
 from mash.filesystem.view import Path
 from mash.shell.shell import build, set_completions, set_functions
-from mash.util import find_fuzzy_matches, has_method, partial_simple
+from mash.shell.function import ShellFunction as Function
+from mash.util import find_fuzzy_matches, hamming, has_method, is_digit, partial_simple
 
 cd_aliasses = 'cd_aliasses'
 path_delimiter = '/'
@@ -27,8 +29,15 @@ class ShellWithFileSystem:
         self.set_shell_completions(cls)
 
         self.shell = cls(save_session_prehook=self.repository.snapshot,
-                         load_session_posthook=self.repository.load)
-        self.shell.set_do_char_method(self.repository.cd, OPTIONS)
+                         load_session_posthook=self.repository.load,
+                         completekey='tab'
+                         )
+
+        for option in OPTIONS:
+            func = partial_simple(self.repository.cd, option)
+            self.shell.add_special_function(option, func)
+
+        self.shell._default_method = self.default_method
 
     def _set_shell_functions(self, cls):
         # convert methods to functions
@@ -67,12 +76,18 @@ class ShellWithFileSystem:
                        }, cls)
 
     def pwd(self):
+        """Print the path to the current working directory
+        """
         return ' '.join(self.repository.full_path)
 
     def get(self, *path: str):
+        """Return the value of the file associated with `path`.
+        """
         return self.repository.get(path)
 
     def set(self, *args: str):
+        """Assign a value to the file k.
+        """
         k, *values = args
 
         if len(values) == 1:
@@ -117,8 +132,9 @@ class ShellWithFileSystem:
 
     def add_cd_alias(self, dirname: str):
         # create alias
-        cd_dirname = partial(self.repository.cd, dirname)
-        cd_dirname.__name__ = f'{self.repository.cd.__name__}({dirname})'
+        func = partial(self.repository.cd, dirname)
+        name = f'{self.repository.cd.__name__}({dirname})'
+        cd_dirname = Function(func, name, f'cd {dirname}')
 
         self.shell.add_functions({dirname: cd_dirname},
                                  group_key=cd_aliasses)
@@ -156,3 +172,18 @@ class ShellWithFileSystem:
                 return results[:1]
 
         return results
+
+    def default_method(self, dirname: str):
+        candidates = self.complete_cd(dirname, None, None, None)
+        if candidates:
+            error = hamming(dirname, candidates[0])
+            if error < 0.9:
+                self.repository.cd(candidates[0])
+                return
+
+        if is_digit(dirname):
+            self.repository.cd(dirname)
+            return
+
+        debug('No matching directory found')
+        return dirname
